import simple_jsonrpc from 'simple-jsonrpc-js';

const API_CLIENT_METHOD_CALLBACK = 'callback';

const API_SERVER_METHOD_AUTHENTICATE = 'authenticate';
const API_SERVER_METHOD_GET_ROOT_OBJECT = 'getRootObject';
const API_SERVER_METHOD_GET_VALUE = 'get';
const API_SERVER_METHOD_SET_VALUE = 'set';
const API_SERVER_METHOD_CALL_FUNCTION = 'execute';
const API_SERVER_METHOD_DESCRIBE_VARIABLES = 'listProperties';
const API_SERVER_METHOD_DESCRIBE_FUNCTIONS = 'listMethods';

const API_VALUE_TYPE_EMPTY = 'empty';
const API_VALUE_TYPE_SCRIPT_OBJECT = 'scriptObject';
const API_VALUE_INVALID_SCRIPT_OBJECT = 0;

function isPlainValueType(value) {
    return ['boolean', 'number', 'string'].includes(typeof value);
}
/**
 * Manages the communication between the client and the Zoiper5 RPC API server.
 */
class RPCManager {
    /**
     * Creates and configures a {@link RPCManager} instance that needs to be initialized via {@link initialize()}.
     */
    constructor(globalCallbacksMap) {
        this._lastScriptObjectId = API_VALUE_INVALID_SCRIPT_OBJECT;
        this._scriptObjectIdMap = new WeakMap();
        const globalCallbackFunctions = Object.values(globalCallbacksMap);
        globalCallbackFunctions.forEach((fn) => this._assignScriptObjectId(fn));
        this._globalCallbacksMap = globalCallbacksMap;
        this._scriptObjectListStack = [globalCallbackFunctions];
        this._jrpc = new simple_jsonrpc();
        this._jrpc.toStream = (message) => {
            this.sendMessage(message);
        };
        this._jrpc.on(API_CLIENT_METHOD_CALLBACK, (serializedTarget, ...serializedArgs) => {
            return this._handleAPICallback(serializedTarget.scriptObject, serializedArgs);
        });
    }
    /**
     * Registers the provided callback list via the {@link RPCManager:constructor()},
     * and starts managing the client-server communication.
     * @param token An API token obtained in advance.
     * @returns An instance of the Zoiper5 API root object - an instance of class `Phone`.
     */
    async initialize(token) {
        await this._jrpc.call(API_SERVER_METHOD_AUTHENTICATE, [token]);
        const response = await this._jrpc.call(API_SERVER_METHOD_GET_ROOT_OBJECT);
        const apiRootObject = await this._deserialize(response);
        for (const [name, fn] of Object.entries(this._globalCallbacksMap)) {
            await apiRootObject.registerCallback(name, fn);
        }
        return apiRootObject;
    }
    /**
     * Send a message from the server to the manager for proccessing.
     * @param message The message that gets sent to the manager.
     */
    processMessage(message) {
        return this._jrpc.messageHandler(message);
    }
    sendMessage() {
        throw new Error('Method "sendMessage" is not implemented');
    }
    _assignScriptObjectId(object) {
        if (!this._scriptObjectIdMap.has(object)) {
            this._lastScriptObjectId += 1;
            this._scriptObjectIdMap.set(object, this._lastScriptObjectId);
        }
    }
    _createGetter(serializedTarget, propertyName) {
        return async () => {
            const response = await this._jrpc.call(API_SERVER_METHOD_GET_VALUE, [
                serializedTarget,
                propertyName,
            ]);
            return this._deserialize(response);
        };
    }
    _createSetter(serializedTarget, propertyName) {
        return async (value) => {
            await this._jrpc.call(API_SERVER_METHOD_SET_VALUE, [
                serializedTarget,
                propertyName,
                this._serialize(value),
            ]);
        };
    }
    _createMethod(serializedTarget, methodName) {
        return (...args) => {
            return this._callAPIMethod(serializedTarget, methodName, args);
        };
    }
    _findScriptObject(scriptObjectId) {
        for (const scriptObjectList of this._scriptObjectListStack) {
            for (const scriptObject of scriptObjectList) {
                if (this._scriptObjectIdMap.get(scriptObject) === scriptObjectId) {
                    return scriptObject;
                }
            }
        }
    }
    async _callAPIMethod(serializedTarget, methodName, args) {
        const scriptObjectList = args.filter((arg) => arg instanceof Object);
        scriptObjectList.forEach((obj) => this._assignScriptObjectId(obj));
        this._scriptObjectListStack.unshift(scriptObjectList);
        try {
            const response = await this._jrpc.call(API_SERVER_METHOD_CALL_FUNCTION, [
                serializedTarget,
                methodName,
                ...args.map((arg) => this._serialize(arg)),
            ]);
            return await this._deserialize(response);
        }
        finally {
            this._scriptObjectListStack.shift();
        }
    }
    async _handleAPICallback(functionId, args) {
        const callbackFunction = this._findScriptObject(functionId);
        if (typeof callbackFunction !== 'function') {
            throw new Error(`Function was not found - ID: ${functionId}`);
        }
        const functionArguments = await Promise.all(args.map((arg) => this._deserialize(arg)));
        await callbackFunction.apply(null, functionArguments);
    }
    async _describeScriptObject(serializedTarget) {
        const scriptObject = {};
        const properties = await this._jrpc.call(API_SERVER_METHOD_DESCRIBE_VARIABLES, [serializedTarget]);
        const methods = await this._jrpc.call(API_SERVER_METHOD_DESCRIBE_FUNCTIONS, [serializedTarget]);
        this._scriptObjectIdMap.set(scriptObject, serializedTarget.scriptObject);
        for (const property of properties) {
            Object.defineProperty(scriptObject, property, {
                enumerable: true,
                get: this._createGetter(serializedTarget, property),
                set: this._createSetter(serializedTarget, property),
            });
        }
        for (const method of methods) {
            Object.defineProperty(scriptObject, method, {
                enumerable: true,
                value: this._createMethod(serializedTarget, method),
            });
        }
        return Object.seal(scriptObject);
    }
    _serialize(value) {
        if (isPlainValueType(value)) {
            return value;
        }
        if (value === undefined) {
            return { type: API_VALUE_TYPE_EMPTY };
        }
        let scriptObjectId;
        if (value === null) {
            scriptObjectId = API_VALUE_INVALID_SCRIPT_OBJECT;
        }
        else {
            scriptObjectId = this._scriptObjectIdMap.get(value);
        }
        if (scriptObjectId === undefined) {
            throw new Error(`Unexpected value: ${value}`);
        }
        return { type: API_VALUE_TYPE_SCRIPT_OBJECT, scriptObject: scriptObjectId };
    }
    async _deserialize(value) {
        if (isPlainValueType(value)) {
            return value;
        }
        if (typeof value === 'object' && value !== null) {
            if (value.type === API_VALUE_TYPE_EMPTY) {
                return undefined;
            }
            if (value.type === API_VALUE_TYPE_SCRIPT_OBJECT) {
                return value.scriptObject === API_VALUE_INVALID_SCRIPT_OBJECT ? null : this._describeScriptObject(value);
            }
        }
        throw new Error(`Unexpected value: ${value}`);
    }
}

class ConnectionManager {
    constructor(rpcManager, options) {
        this._rpcManager = rpcManager;
        this._options = options;
    }
    openConnection() {
        return new Promise((resolve, reject) => {
            this._socket = new WebSocket(this._options.url, 'text');
            this._socket.onmessage = (event) => {
                this._rpcManager.processMessage(event.data);
            };
            this._rpcManager.sendMessage = (message) => {
                this._socket.send(message);
            };
            this._socket.onclose = (event) => {
                var _a;
                (_a = this.onClose) === null || _a === void 0 ? void 0 : _a.call(this, event);
            };
            this._socket.onopen = () => resolve();
            this._socket.onerror = reject;
        });
    }
    closeConnection() {
        var _a;
        (_a = this._socket) === null || _a === void 0 ? void 0 : _a.close();
    }
}

export { ConnectionManager, RPCManager };
//# sourceMappingURL=zoiper5-rpc-api-js-wrapper.esm.js.map
